from typing import TYPE_CHECKING, TypedDict, Literal
from os import remove
from os.path import exists, join

__all__ = ["CompressionOptions", "setup_dataframe_serialization", "compression_options", "compression_suffix"]

SUPPORTED_COMPRESSION_METHODS = ["zip", "gzip", "bz2", "zstd", "xz", "tar"]
class CompressionOptions(TypedDict):
    """Options for compression methods used in serialization.
    """

    method: Literal["zip", "gzip", "bz2", "zstd", "xz", "tar"] | None
    """The compression method to use. If None, no compression is applied.
    Supported methods are: zip, gzip, bz2, zstd, xz, tar.
    If use zstd, please install zstandard library otherwise it will raise ImportError.
    For more information or other compression options, see https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.to_pickle.html
    """

if TYPE_CHECKING:
    from .context import Context


def setup_dataframe_serialization(context: 'Context', compression: CompressionOptions | None = None) -> None:
    """
    Setup the DataFrame serialization for the compression operation. This function needs to be called before using DataFrame serialization. 
    This function ensures that DataFrames are serialized to a special file based on the specified compression method.
    The configuration is stored in the `__compression_options.json` file in the pkg_data_dir. So it can be persisted across sessions
    and stored in the session directory for later retrieval.
    """
    try:
        if compression is None:
            if exists(join(context.pkg_data_dir, COMPRESSION_OPTIONS_FILE)):
                remove(join(context.pkg_data_dir, COMPRESSION_OPTIONS_FILE))
            return
        elif compression["method"] not in SUPPORTED_COMPRESSION_METHODS:
            raise ValueError(f"Unsupported compression method: {compression['method']}. Supported methods are: {SUPPORTED_COMPRESSION_METHODS}.")
        elif compression is not None and compression["method"] == "zstd":
            # If zstd compression is specified, ensure that the zstandard library is available.
            try:
                import zstandard as zstd  # type: ignore
            except ImportError:
                raise ImportError("To use zstd compression, please install the zstandard library.")
        
        # write compression options to a file if exist then overwrite it
        with open(join(context.pkg_data_dir, COMPRESSION_OPTIONS_FILE), "w") as f:
            import json
            json.dump(compression, f)
        
    except ImportError:
        raise ImportError("To Setup DataFrame serialization, pandas is required. Please install it with `poetry install pandas`.")

COMPRESSION_OPTIONS_FILE = "__compression_options.json"

def compression_options(context: 'Context') -> CompressionOptions | None:
    """
    Retrieve the compression options from the session directory.
    If no options are set, return None.
    """
    import json
    try:
        with open(join(context.pkg_data_dir, COMPRESSION_OPTIONS_FILE), "r") as f:
            return json.load(f)
    except FileNotFoundError:
        return None
    except json.JSONDecodeError:
        print("Error decoding JSON from compression options file. Returning None.")
        return None
    except Exception as e:
        print(f"An unexpected error occurred while reading compression options: {e}. Returning None.")
        return None
    
def compression_suffix(context: 'Context') -> str:
    """
    Get the file suffix based on the compression method.
    If no compression is specified, return an empty string.
    """
    compression = compression_options(context)

    if compression is None or compression["method"] is None:
        return ".pkl"
    
    method = compression["method"]
    if method == "zip":
        return ".zip"
    elif method == "gzip":
        return ".gz"
    elif method == "bz2":
        return ".bz2"
    elif method == "zstd":
        return ".zst"
    elif method == "xz":
        return ".xz"
    elif method == "tar":
        return ".tar"
    else:
        return ".pkl"  # Default case if method is not recognized


import unittest
from oocana import data
from json import dumps as json_dumps

class TestData(unittest.TestCase):

    def test_store_key_missing(self):
        store_key_dict = {
            "executor": "executor",
            "handle": "handle",
            "job_id": "job_id",
        }

        with self.assertRaises(ValueError):
            data.StoreKey(**store_key_dict)

    def test_store_key_extra(self):
        store_key_dict = {
            "executor": "executor",
            "handle": "handle",
            "job_id": "job_id",
            "session_id": "session_id",
            "extra": "extra"
        }
        store_key = data.StoreKey(**store_key_dict)
        self.assertEqual(store_key.executor, 'executor')
        self.assertEqual(store_key.handle, 'handle')
        self.assertEqual(store_key.job_id, 'job_id')
        self.assertEqual(store_key.session_id, 'session_id')

    def test_block_info_missing(self):
        block_info_dict = {
            "session_id": "session_id_one",
            "stacks": ["stack1", "stack2"],
        }

        with self.assertRaises(ValueError):
            data.BlockInfo(**block_info_dict)

    def test_block_info_extra(self):
        block_info_dict = {
            "session_id": "session_id_one",
            "job_id": "job_id_one",
            "stacks": ["stack1", "stack2"],
            "block_path": "block_path_one",
            "extra": "extra"
        }

        block_info = data.BlockInfo(**block_info_dict)
        self.assertEqual(block_info.session_id, "session_id_one")
        self.assertEqual(block_info.job_id, "job_id_one")
        self.assertEqual(block_info.stacks, ["stack1", "stack2"])
        self.assertEqual(block_info.block_path, "block_path_one")

    def test_dataclass_dumps(self):
        block_info_dict = {
            "session_id": "session_id_one",
            "job_id": "job_id_one",
            "stacks": ["stack1", "stack2"],
            "block_path": "block_path_one",
            "extra": "extra"
        }

        block_info = data.BlockInfo(**block_info_dict)
        serialize_block_info = data.dumps(block_info)
        self.assertEqual(serialize_block_info, '{"session_id": "session_id_one", "job_id": "job_id_one", "stacks": ["stack1", "stack2"], "block_path": "block_path_one"}')

        list_serialize_block_info = data.dumps([block_info])
        self.assertEqual(list_serialize_block_info, '[{"session_id": "session_id_one", "job_id": "job_id_one", "stacks": ["stack1", "stack2"], "block_path": "block_path_one"}]')

        key_serialize_block_info = data.dumps({"key": block_info})
        self.assertEqual(key_serialize_block_info, '{"key": {"session_id": "session_id_one", "job_id": "job_id_one", "stacks": ["stack1", "stack2"], "block_path": "block_path_one"}}')

        with self.assertRaises(TypeError):
            json_dumps(block_info)

    def test_dataclass_dumps_with_none(self):
        block_info_dict = {
            "session_id": "session_id_one",
            "job_id": "job_id_one",
            "stacks": ["stack1", "stack2"],
            "block_path": None,
            "extra": "extra"
        }

        block_info = data.BlockInfo(**block_info_dict)
        serialize_block_info = data.dumps(block_info.block_dict())
        self.assertEqual(serialize_block_info, '{"session_id": "session_id_one", "job_id": "job_id_one", "stacks": ["stack1", "stack2"]}')

        with self.assertRaises(TypeError):
            json_dumps(block_info)


import asyncio
from dataclasses import asdict
from .data import BlockInfo, StoreKey, JobDict, BlockDict, BinValueDict, VarValueDict
from .mainframe import Mainframe
from .handle import HandleDef, OutputHandleDef
from typing import Dict, Any, TypedDict, Optional, Callable, Mapping
from types import MappingProxyType
from base64 import b64encode
from io import BytesIO
from .throttler import throttle
from .preview import PreviewPayload, DataFrame, PreviewPayloadInternal, ShapeDataFrame
from .data import EXECUTOR_NAME
import os.path
import logging
import random
import string
import hashlib

__all__ = ["Context", "HandleDefDict", "BlockJob", "BlockExecuteException"]

def random_string(length=8):
    return ''.join(random.choices(string.ascii_letters + string.digits, k=length))

def string_hash(text: str) -> str:
    """
    Generates a deterministic hash for a given string.
    """
    return hashlib.sha256(text.encode('utf-8')).hexdigest()

class ToNode(TypedDict):
    node_id: str

    input_handle: str

class ToFlow(TypedDict):
    output_handle: str

class BlockExecuteException(Exception):
    """Exception raised when a block execution fails."""
    def __init__(self, message: str):
        super().__init__(message)
        self.message = message

    def __str__(self):
        return f"BlockExecuteException: {self.message}"

class BlockJob:

    __outputs_callbacks: set[Callable[[Dict[str, Any]], None]]
    __finish_future: asyncio.Future[None]

    def __init__(self, outputs_callbacks: set[Callable[[Dict[str, Any]], None]], future: asyncio.Future[None]) -> None:
        self.__outputs_callbacks = outputs_callbacks
        self.__finish_future = future

    def add_output_callback(self, fn: Callable[[Dict[str, Any]], None]):
        """Register a callback function to handle the output of the block.
        :param fn: the callback function, it should accept a dict as the first parameter, the output handle will be the key, and the output value will be the value.

        Note: the callback function is running in mqtt's callback thread, so it currently does not support run context.sendXX message directly.
        """
        if not callable(fn):
            raise ValueError("output_callback should be a callable function.")
        self.__outputs_callbacks.add(fn)

    def finish(self) -> asyncio.Future[None]:
        """Wait for the block to finish and return the future.
        
        Returns:
            asyncio.Future: A future that will be resolved when the block finishes.
            If the block finishes with an error, the future will throw a Exception.
        """
        return self.__finish_future

class HandleDefDict(TypedDict):
    """a dict that represents the handle definition, used in the block schema output and input defs.
    """

    handle: str
    """the handle of the output, should be defined in the block schema output defs, the field name is handle
    """

    description: str | None
    """the description of the output, should be defined in the block schema output defs, the field name is description
    """

    json_schema: Dict[str, Any] | None
    """the schema of the output, should be defined in the block schema output defs, the field name is json_schema
    """

    kind: str | None
    """the kind of the output, should be defined in the block schema output defs, the field name is kind
    """

    nullable: bool
    """if the output can be None, should be defined in the block schema output defs, the field name is nullable
    """

    is_additional: bool
    """if the output is an additional output, should be defined in the block schema output defs, the field name is is_additional
    """

class QueryBlockResponse(TypedDict):
    description: str | None
    """the description of the block, if the block has no description, this field should be
    None.
    """

    inputs_def: Dict[str, HandleDefDict] | None
    """the inputs of the block, should be a dict, if the block has no inputs
    this field should be None.
    """

    outputs_def: Dict[str, HandleDefDict] | None
    """the outputs of the block, should be a dict, if the block has no outputs  
    this field should be None.
    """

    additional_outputs: bool
    """if the block author declare this block accept additional outputs, this field should be True, otherwise False.
    """

    additional_inputs: bool
    """if the block author declare this block accept additional inputs, this field should be True, otherwise False.
    """


class FlowDownstream(TypedDict):
    output_handle: str
    output_handle_def: HandleDefDict | None

class NodeDownstream(TypedDict):
    node_id: str
    description: str | None
    """node description"""
    input_handle: str
    input_handle_def: HandleDefDict | None

class Downstream(TypedDict):
    to_flow: list[FlowDownstream]
    to_node: list[NodeDownstream]

class OnlyEqualSelf:
    def __eq__(self, value: object) -> bool:
        return self is value

class OOMOL_LLM_ENV(TypedDict):
    base_url: str
    """{basUrl}/v1 openai compatible endpoint
    """
    base_url_v1: str
    api_key: str
    models: list[str]

class HostInfo(TypedDict):
    gpu_vendor: str
    gpu_renderer: str

class Context:
    __inputs: Dict[str, Any]

    __block_info: BlockInfo
    __outputs_def: Dict[str, OutputHandleDef]
    # Only dict can support some field type like `Optional[FieldSchema]`(this key can not in dict). Dataclass will always convert it to None if the field is not set which will cause some issues.
    __outputs_def_dict: Dict[str, HandleDefDict]
    __inputs_def: Dict[str, HandleDefDict]
    __store: Any
    __keep_alive: OnlyEqualSelf = OnlyEqualSelf()
    __session_dir: str
    __tmp_dir: str
    __package_name: str | None = None
    _logger: Optional[logging.Logger] = None
    __pkg_data_dir: str

    # TODO: remove the pkg_dir parameter, use pkg_data_dir instead.
    def __init__(
        self, *, inputs: Dict[str, Any], blockInfo: BlockInfo, mainframe: Mainframe, store, inputs_def, outputs_def: Dict[str, Any], session_dir: str, tmp_dir: str, package_name: str, pkg_dir: str
    ) -> None:

        self.__block_info = blockInfo

        self.__mainframe = mainframe
        self.__store = store
        self.__inputs = inputs

        self.__outputs_def_dict = outputs_def
        outputs_defs_cls = {}
        if outputs_def is not None:
            for k, v in outputs_def.items():
                outputs_defs_cls[k] = OutputHandleDef(**v)
        self.__outputs_def = outputs_defs_cls
        self.__inputs_def = inputs_def
        self.__session_dir = session_dir
        self.__tmp_dir = tmp_dir
        self.__package_name = package_name
        self.__pkg_data_dir = pkg_dir

    @property
    def logger(self) -> logging.Logger:
        """a custom logger for the block, you can use it to log the message to the block log. this logger will report the log by context report_logger api.
        """

        # setup after init, so the logger always exists
        if self._logger is None:
            raise ValueError("logger is not setup, please setup the logger in the block init function.")
        return self._logger

    @property
    def session_dir(self) -> str:
        """a temporary directory for the current session, all blocks in the one session will share the same directory.
        """
        return self.__session_dir
    
    @property
    def tmp_dir(self) -> str:
        """a temporary directory for the current follow, all blocks in the this flow will share the same directory. this directory will be cleaned if this session finish successfully, otherwise it will be kept for debugging or other purpose.
        """
        return self.__tmp_dir
    
    @property
    def tmp_pkg_dir(self) -> str:
        """a temporary directory for the current package, all blocks in the this package will share the same directory. this directory will be cleaned if this session finish successfully, otherwise it will be kept for debugging or other purpose.
        """
        return os.path.join(self.__tmp_dir, self.__package_name) if self.__package_name else self.__tmp_dir

    @property
    def pkg_dir(self) -> str:
        """Deprecated, use pkg_data_dir instead.
        """
        return self.__pkg_data_dir

    @property
    def pkg_data_dir(self) -> str:
        """A directory for the current package data, all blocks in this package will share the same directory. 
        This directory's content will be persisted after the session finishes, so you can use it to store some data that need to be shared between blocks in the same package.
        Please note that this directory is not cleaned after the session finishes, so you need to manage the content of this directory by yourself. The same package with different versions will share the same directory, so you need to be careful with different package versions.
        """
        return self.__pkg_data_dir

    @property
    def keepAlive(self):
        return self.__keep_alive

    @property
    def inputs(self):
        return self.__inputs
    
    @property
    def inputs_def(self) -> Mapping[str, HandleDefDict]:
        """a dict that represents the input definitions, used in the block schema input defs.
        This is a read-only property, you can not modify it.
        """
        return MappingProxyType(self.__inputs_def) if self.__inputs_def is not None else MappingProxyType({})

    @property
    def outputs_def(self) -> Mapping[str, HandleDefDict]:
        """a dict that represents the output definitions, used in the block schema output defs.
        This is a read-only property, you can not modify it.
        """
        return MappingProxyType(self.__outputs_def_dict) if self.__outputs_def_dict is not None else MappingProxyType({})

    @property
    def session_id(self):
        return self.__block_info.session_id

    @property
    def job_id(self):
        return self.__block_info.job_id
    
    @property
    def job_info(self) -> JobDict:
        return self.__block_info.job_info()
    
    @property
    def block_info(self) -> BlockDict:
        return self.__block_info.block_dict()
    
    @property
    def node_id(self) -> str:
        return self.__block_info.stacks[-1].get("node_id", None)
    
    @property
    def oomol_llm_env(self) -> OOMOL_LLM_ENV:
        """this is a dict contains the oomol llm environment variables
        """

        oomol_llm_env: OOMOL_LLM_ENV = {
            "base_url": os.getenv("OOMOL_LLM_BASE_URL", ""),
            "base_url_v1": os.getenv("OOMOL_LLM_BASE_URL_V1", ""),
            "api_key": os.getenv("OOMOL_LLM_API_KEY", ""),
            "models": os.getenv("OOMOL_LLM_MODELS", "").split(","),
        }

        for key, value in oomol_llm_env.items():
            if value == "" or value == []:
                self.send_warning(
                    f"OOMOL_LLM_ENV variable {key} is ({value}), this may cause some features not working properly."
                )

        return oomol_llm_env

    @property
    def host_info(self) -> HostInfo:
        """this is a dict contains the host information
        """
        return {
            "gpu_vendor": os.getenv("OOMOL_HOST_GPU_VENDOR", "unknown"),
            "gpu_renderer": os.getenv("OOMOL_HOST_GPU_RENDERER", "unknown"),
        }

    @property
    def host_endpoint(self) -> str | None:
        """A host endpoint that allows containers to access services running on the host system.
        
        Returns:
            str: The host endpoint if available.
            None: If the application is running in a cloud environment where no host endpoint is defined.
        """
        return os.getenv("OO_HOST_ENDPOINT", None)

    def __store_ref(self, handle: str):
        return StoreKey(
            executor=EXECUTOR_NAME,
            handle=handle,
            job_id=self.job_id,
            session_id=self.session_id,
        )
    
    def __is_basic_type(self, value: Any) -> bool:
        return isinstance(value, (int, float, str, bool))
    
    def __wrap_output_value(self, handle: str, value: Any):
        """
        wrap the output value:
        if the value is a var handle, store it in the store and return the reference.
        if the value is a bin handle, store it in the store and return the reference.
        if the handle is not defined in the block outputs schema, raise an ValueError.
        otherwise, return the value.
        :param handle: the handle of the output
        :param value: the value of the output
        :return: the wrapped value
        """
        # __outputs_def should never be None
        if self.__outputs_def is None:
            return value
        
        output_def = self.__outputs_def.get(handle)
        if output_def is None:
            raise ValueError(
                f"Output handle key: [{handle}] is not defined in Block outputs schema."
            )
        
        if output_def.is_var_handle() and not self.__is_basic_type(value):
            ref = self.__store_ref(handle)
            self.__store[ref] = value

            serialize_path = None
            # only cache root flow
            if len(self.__block_info.stacks) < 2 and output_def.need_serialize_var_for_cache() and value.__class__.__name__ == 'DataFrame' and callable(getattr(value, 'to_pickle', None)):
                from .serialization import compression_suffix, compression_options
                suffix = compression_suffix(context=self)
                compression = compression_options(context=self)
                flow_node = self.__block_info.stacks[-1].get("flow", "unknown") + "-" + self.node_id
                serialize_path = f"{self.pkg_data_dir}/.cache/{string_hash(flow_node)}/{handle}{suffix}"
                os.makedirs(os.path.dirname(serialize_path), exist_ok=True)
                try:
                    copy_value = value.copy()  # copy the value to avoid blocking the main thread
                    import threading
                    def write_pickle():
                        copy_value.to_pickle(serialize_path, compression=compression)
                    thread = threading.Thread(target=write_pickle)
                    thread.start()
                except IOError as e:
                    pass
            var: VarValueDict = {
                "__OOMOL_TYPE__": "oomol/var",
                "value": asdict(ref),
                "serialize_path": serialize_path,
            }
            return var
        
        if output_def.is_bin_handle():
            if not isinstance(value, bytes):
                self.send_warning(
                    f"Output handle key: [{handle}] is defined as binary, but the value is not bytes."
                )
                return value
            
            bin_file = f"{self.session_dir}/binary/{self.session_id}/{self.job_id}/{handle}"
            os.makedirs(os.path.dirname(bin_file), exist_ok=True)
            try:
                with open(bin_file, "wb") as f:
                    f.write(value)
            except IOError as e:
                raise IOError(
                    f"Output handle key: [{handle}] is defined as binary, but an error occurred while writing the file: {e}"
                )

            if os.path.exists(bin_file):
                bin_value: BinValueDict = {
                    "__OOMOL_TYPE__": "oomol/bin",
                    "value": bin_file,
                }
                return bin_value
            else:
                raise IOError(
                    f"Output handle key: [{handle}] is defined as binary, but the file is not written."
                )
        return value

    def output(self, key: str, value: Any, *, to_node: list[ToNode] | None = None, to_flow: list[ToFlow] | None = None):
        """
        output the value to the next block

        :param key: str, the key of the output, should be defined in the block schema output defs, the field name is handle
        :param value: Any, the value of the output
        :param to_node: list[ToNode] | None, the target node(with input handle) to send the output
        :param to_flow: list[ToFlow] | None, the target flow(with output handle) to send the output
        if both to_node and to_flow are None, the output will be sent to all connected nodes and flows.
        """

        try:
            wrap_value = self.__wrap_output_value(key, value)
        except ValueError as e:
            self.send_warning(
                f"{e}"
            )
            return
        except IOError as e:
            self.send_warning(
                f"{e}"
            )
            return

        if to_node is not None or to_flow is not None:
            target = {
                "to_node": to_node,
                "to_flow": to_flow,
            }
        else:
            target = None

        payload = {
            "type": "BlockOutput",
            "handle": key,
            "output": wrap_value,
        }
        if target is not None:
            payload["options"] = {"target": target}
        self.__mainframe.send(self.job_info, payload)
    
    def outputs(self, outputs: Dict[str, Any]):
        """
        output the value to the next block

        map: Dict[str, Any], the key of the output, should be defined in the block schema output defs, the field name is handle
        """

        values = {}
        for key, value in outputs.items():
            try:
                wrap_value = self.__wrap_output_value(key, value)
                values[key] = wrap_value
            except ValueError as e:
                self.send_warning(
                    f"{e}"
                )
            except IOError as e:
                self.send_warning(
                    f"{e}"
                )
        self.__mainframe.send(self.job_info, {
            "type": "BlockOutputs",
            "outputs": values,
        })

        

    def finish(self, *, result: Dict[str, Any] | None = None, error: str | None = None):
        """
        finish the block, and send the result to oocana.
        if error is not None, the block will be finished with error.
        then if result is not None, the block will be finished with result.
        lastly, if both error and result are None, the block will be finished without any result.
        """

        if error is not None:
            self.__mainframe.send(self.job_info, {"type": "BlockFinished", "error": error})
        elif result is not None:
            wrap_result = {}
            if isinstance(result, dict):
                for key, value in result.items():
                    try:
                        wrap_result[key] = self.__wrap_output_value(key, value)
                    except ValueError as e:
                        self.send_warning(
                            f"Output handle key: [{key}] is not defined in Block outputs schema. {e}"
                        )
                    except IOError as e:
                        self.send_warning(
                            f"Output handle key: [{key}] is not defined in Block outputs schema. {e}"
                        )

                self.__mainframe.send(self.job_info, {"type": "BlockFinished", "result": wrap_result})
            else:
                raise ValueError(
                    f"result should be a dict, but got {type(result)}"
                )
        else:
            self.__mainframe.send(self.job_info, {"type": "BlockFinished"})

    def send_message(self, payload):
        """
        send a message to the block, this message will be displayed in the log of the block.
        :param payload: the payload of the message, it can be a string or a dict
        """
        self.__mainframe.report(
            self.block_info,
            {
                "type": "BlockMessage",
                "payload": payload,
            },
        )
    
    def __dataframe(self, payload: PreviewPayload) -> PreviewPayloadInternal:
        target_dir = os.path.join(self.tmp_dir, self.job_id)
        os.makedirs(target_dir, exist_ok=True)
        csv_file = os.path.join(target_dir, f"{random_string(8)}.csv")
        if isinstance(payload, DataFrame):
            payload.to_csv(path_or_buf=csv_file)
            payload = { "type": "table", "data": csv_file }

        if isinstance(payload, dict) and payload.get("type") == "table":
            df = payload.get("data")
            if isinstance(df, ShapeDataFrame):
                df.to_csv(path_or_buf=csv_file)
                payload = { "type": "table", "data": csv_file }
            else:
                print("dataframe is not support shape property")
        
        return payload

    def __matplotlib(self, payload: PreviewPayloadInternal) -> PreviewPayloadInternal:
        # payload is a matplotlib Figure
        if hasattr(payload, 'savefig'):
            fig: Any = payload
            buffer = BytesIO()
            fig.savefig(buffer, format='png')
            buffer.seek(0)
            png = buffer.getvalue()
            buffer.close()
            url = f'data:image/png;base64,{b64encode(png).decode("utf-8")}'
            payload = { "type": "image", "data": url }

        return payload
        

    def preview(self, payload: PreviewPayload, id: str | None = None):
        payload_internal = self.__dataframe(payload)
        payload_internal = self.__matplotlib(payload_internal)

        if id is not None:
            payload_internal["id"] = id #type: ignore

        self.__mainframe.report(
            self.block_info,
            {
                "type": "BlockPreview",
                "payload": payload_internal,  # type: ignore
            },
        )

    @throttle(0.3)
    def report_progress(self, progress: float | int):
        """report progress

        This api is used to report the progress of the block. but it just effect the ui progress not the real progress.
        This api is throttled. the minimum interval is 0.3s. 
        When you first call this api, it will report the progress immediately. After it invoked once, it will report the progress at the end of the throttling period.

        |       0.25 s        |   0.2 s  |
        first call       second call    third call  4 5 6 7's calls
        |                     |          |          | | | |
        | -------- 0.3 s -------- | -------- 0.3 s -------- |
        invoke                  invoke                    invoke
        :param float | int progress: the progress of the block, the value should be in [0, 100].
        """
        self.__mainframe.report(
            self.block_info,
            {
                "type": "BlockProgress",
                "rate": progress,
            }
        )

    def report_log(self, line: str, stdio: str = "stdout"):
        self.__mainframe.report(
            self.block_info,
            {
                "type": "BlockLog",
                "log": line,
                stdio: stdio,
            },
        )

    def log_json(self, payload):
        self.__mainframe.report(
            self.block_info,
            {
                "type": "BlockLogJSON",
                "json": payload,
            },
        )

    def send_warning(self, warning: str):
        self.__mainframe.report(self.block_info, {"type": "BlockWarning", "warning": warning})

    def send_error(self, error: str):
        '''
        deprecated, use error(error) instead.
        consider to remove in the future.
        '''
        self.error(error)

    def error(self, error: str):
        self.__mainframe.send(self.job_info, {"type": "BlockError", "error": error})

    async def query_downstream(self, handles: list[str] | None = None) -> Dict[str, Downstream]:
        """
        query the downstream nodes of the given output handles.
        :param handle: the handle of the output, should be defined in the block schema output defs. If None means query all handles.
        :return: a dict that contains the downstream nodes, including the node id and input handle.
        """
        request_id = random_string(16)
        loop = asyncio.get_running_loop()
        f: asyncio.Future[Dict[str, Any]] = loop.create_future()

        def response_callback(payload: Dict[str, Any]):
            if payload.get("request_id") != request_id:
                return
            self.__mainframe.remove_request_response_callback(self.session_id, request_id, response_callback)
            if payload.get("result") is not None:
                loop.call_soon_threadsafe(lambda: f.set_result(payload.get("result", {})))
            elif payload.get("error") is not None:
                loop.call_soon_threadsafe(lambda: f.set_exception(ValueError(payload.get("error", "Unknown error occurred while querying the downstream."))))

        self.__mainframe.add_request_response_callback(self.session_id, request_id, response_callback)

        self.__mainframe.send(self.job_info, {
            "type": "BlockRequest",
            "action": "QueryDownstream",
            "handles": handles,
            "session_id": self.session_id,
            "job_id": self.job_id,
            "request_id": request_id,
        })

        return await f
    
    async def query_block(self, block: str) -> QueryBlockResponse:
        """
        this is a experimental api, it is used to query the block information..

        query a block by its id.
        :param block: the id of the block to query. format: `self::<block_name>` or `<package_name>::<block_name>`.
        :return: a dict that contains the block information, including the block schema, inputs and outputs.

        if the block is not found, it will raise a ValueError.

        example:
        ```python
        response = await context.query_block("self::my_block")
        print(response)
        """

        request_id = random_string(16)
        loop = asyncio.get_running_loop()
        f: asyncio.Future[QueryBlockResponse] = loop.create_future()

        def response_callback(payload: Dict[str, Any]):
            """
            This callback is called when the block information is received.
            It will return the block information to the caller.
            """
            if payload.get("request_id") != request_id:
                return
            self.__mainframe.remove_request_response_callback(self.session_id, request_id, response_callback)
            
            if payload.get("result") is not None:
                loop.call_soon_threadsafe(lambda: f.set_result(payload.get("result", {})))
            elif payload.get("error") is not None:
                loop.call_soon_threadsafe(lambda: f.set_exception(ValueError(payload.get("error", "Unknown error occurred while querying the block."))))

        
        self.__mainframe.add_request_response_callback(self.session_id, request_id, response_callback)

        self.__mainframe.send(self.job_info, {
            "type": "BlockRequest",
            "action": "QueryBlock",
            "block": block,
            "session_id": self.session_id,
            "job_id": self.job_id,
            "request_id": request_id,
        })

        return await f
        

    def run_block(self, block: str, *, inputs: Dict[str, Any], additional_inputs_def: list[HandleDefDict] | None = None, additional_outputs_def: list[HandleDefDict] | None = None, strict: bool = False) -> BlockJob:
        """
        :param block: the id of the block to run. format: `self::<block_name>` or `<package_name>::<block_name>`.
        :param inputs: the inputs of the block. if the block has no inputs, this parameter can be dict. 
                    If the inputs missing some required inputs, the response's finish future will send {"error": "<error message>" }.
                    some missing inputs will be filled:
                        1. with the default value if the block's input_def has a default value (which is defined in the value field).
                        2. input_def's nullable is true, the missing input will be filled with Null.
        :param additional_inputs_def: additional inputs definitions, this is a list of dicts, each dict should contain the handle(required), description, json_schema, kind, nullable and is_additional fields. This is used to define additional inputs that are not defined in the block schema.
        :param additional_outputs_def: additional outputs definitions, this is a list of dicts, each dict should contain the handle(required), description, json_schema, kind, nullable and is_additional fields. This is used to define additional outputs that are not defined in the block schema.
        :param strict: if True, oocana will use input_def's json_schema(only when it exists) to validate the inputs and if they are not valid oocana will reject to run this block, return error message. otherwise it will run with the inputs as they are.
        :return: a RunResponse object, which contains the event callbacks and output callbacks. You can use the `add_event_callback` and `add_output_callback` methods to register callbacks for the events and outputs of the block. You can also use the `finish` method to wait for the block to finish and get the result.
        Notice do not call any context.send_message or context.report_progress or context.preview and other context methods(which will send message) directly in the callbacks, it may cause deadlock.

        this is a experimental api, it is used to run a block in the current context.
        It will send a request to the mainframe to run the block with the given inputs.
        It will return a RunResponse object, which contains the event callbacks and output callbacks.
        You can use the `add_event_callback` and `add_output_callback` methods to register callbacks for the events and outputs of the block.
        You can also use the `finish` method to wait for the block to finish and get the result or error.

        example:
        ```python
        response = context.run_block("self::my_block", {"input1": "value1", "input2": "value2"})
        response.add_event_callback(lambda event: print(f"Event received: {event}"))
        response.add_output_callback(lambda handle, value: print(f"Output received: {handle} = {value}"))
        payload = await response.finish()
        if payload.get("error"):
            print(f"Block finished with error: {payload['error']}")
        else:
            print(f"Block finished with result: {payload['result']}")
        ```
        """

        # consider use uuid, remove job_id and block_job_id.
        block_job_id = f"{self.job_id}-{block}-{random_string(8)}"
        request_id = random_string(16)
        self.__mainframe.send(self.job_info, {
            "type": "BlockRequest",
            "action": "RunBlock",
            "block": block,
            "block_job_id": block_job_id,
            "payload": {
                "inputs": inputs,
                "additional_inputs_def": additional_inputs_def,
                "additional_outputs_def": additional_outputs_def,
            },
            "stacks": self.__block_info.stacks,
            "strict": strict,
            "request_id": request_id,
        })

        outputs_callbacks: set[Callable[[Dict[str, Any]], None]] = set()

        # run_block will always run in a coroutine, so we can use asyncio.Future to wait for the result.
        loop = asyncio.get_running_loop()
        future: asyncio.Future[None] = loop.create_future()

        def response_callback(payload: Dict[str, Any]):
            """
            This callback is called when an error occurs while running a block.
            It will call the error callbacks registered by the user.
            """
            if payload.get("request_id") != request_id:
                return

            error = payload.get("error")
            # only handle error 
            if error is not None:
                set_future_and_clean(error)

        def run_output_callback(payload: Dict[str, Any]):
            # TODO: run callback in different coroutine, so it can send mqtt message directly.
            for output_callback in outputs_callbacks:
                output_callback(payload)

        def event_callback(payload: Dict[str, Any]):

            if payload.get("session_id") != self.session_id:
                return

            if payload.get("job_id") != block_job_id:
                return
            elif payload.get("type") == "ExecutorReady" or payload.get("type") == "BlockReady" or payload.get("type") == "BlockRequest":
                # ignore these messages
                return

            if payload.get("type") == "BlockOutput":
                output = {}
                output[payload.get("handle")] = payload.get("output")
                run_output_callback(output)
            elif payload.get("type") == "BlockOutputs":
                run_output_callback(payload.get("outputs", {}))
            elif payload.get("type") == "SubflowBlockOutput":
                output = {}
                output[payload.get("handle")] = payload.get("output")
                run_output_callback(output)
            elif payload.get("type") == "SubflowBlockFinished":
                error = payload.get("error")
                set_future_and_clean(error)
            elif payload.get("type") == "BlockFinished":
                result = payload.get("result", {})
                error = payload.get("error")
                if result is not None and not isinstance(result, dict):
                    pass
                elif result is not None:
                    run_output_callback(result)

                set_future_and_clean(error)

        job = BlockJob(outputs_callbacks, future)

        def set_future_and_clean(error: None | str = None):
            self.__mainframe.remove_report_callback(event_callback)
            self.__mainframe.remove_request_response_callback(self.session_id, request_id, response_callback)

            def set_future():
                if future.done():
                    return
                if error is None:
                    future.set_result(None)
                else:
                    future.set_exception(BlockExecuteException(f"run block {block} failed: {error}"))

            loop.call_soon_threadsafe(set_future)
            outputs_callbacks.clear()


        self.__mainframe.add_report_callback(event_callback)
        self.__mainframe.add_request_response_callback(self.session_id, request_id, response_callback)

        return job



from dataclasses import dataclass, asdict
from typing import TypedDict, Literal
from simplejson import JSONEncoder
import simplejson as json

EXECUTOR_NAME = "python"

__all__ = ["dumps", "BinValueDict", "VarValueDict", "JobDict", "BlockDict", "StoreKey", "BlockInfo", "EXECUTOR_NAME", "JobDict", "BinValueDict", "VarValueDict"]

def dumps(obj, **kwargs):
    return json.dumps(obj, cls=DataclassJSONEncoder, ignore_nan=True, **kwargs)

class DataclassJSONEncoder(JSONEncoder):
    def default(self, o): # pyright: ignore[reportIncompatibleMethodOverride]
        if hasattr(o, '__dataclass_fields__'):
            return asdict(o)
        return JSONEncoder.default(self, o)

class BinValueDict(TypedDict):
    value: str
    __OOMOL_TYPE__: Literal["oomol/bin"]

class VarValueDict(TypedDict):
    value: dict
    __OOMOL_TYPE__: Literal["oomol/var"]
    serialize_path: str | None # better to use NotRequired here, but it is not supported in python 3.10 and type_extensions requires high version, For compatibility, use None instead of NotRequired.

class JobDict(TypedDict):
    session_id: str
    job_id: str

class BlockDict(TypedDict):
    session_id: str
    job_id: str
    stacks: list
    block_path: str | None # better to use NotRequired here, but it is not supported in python 3.10 and type_extensions requires high version, For compatibility, use None instead of NotRequired.

# dataclass 默认字段必须一一匹配
# 如果多一个或者少一个字段，就会报错。
# 这里想兼容额外多余字段，所以需要自己重写 __init__ 方法，忽略处理多余字段。同时需要自己处理缺少字段的情况。
@dataclass(frozen=True, kw_only=True)
class StoreKey:
    executor: str
    handle: str
    job_id: str
    session_id: str

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            object.__setattr__(self, key, value)
        for key in self.__annotations__.keys():
            if key not in kwargs:
                raise ValueError(f"missing key {key}")


# 发送 reporter 时，固定需要的 block 信息参数
@dataclass(frozen=True, kw_only=True)
class BlockInfo:

    session_id: str
    job_id: str
    stacks: list
    block_path: str | None = None

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            object.__setattr__(self, key, value)
        for key in self.__annotations__.keys():
            if key not in kwargs and key != "block_path":
                raise ValueError(f"missing key {key}")

    def job_info(self) -> JobDict:
        return {"session_id": self.session_id, "job_id": self.job_id}

    def block_dict(self) -> BlockDict:
        if self.block_path is None:
            return {
                "session_id": self.session_id,
                "job_id": self.job_id,
                "stacks": self.stacks,
            } # type: ignore[return-value]

        return {
            "session_id": self.session_id,
            "job_id": self.job_id,
            "stacks": self.stacks,
            "block_path": self.block_path,
        }
    




import unittest
from oocana import HandleDef, InputHandleDef, OutputHandleDef
from typing import Any, Optional, TypedDict

missing_handle = {
}

redundant_handle = {
    "handle": "test",
    "json_schema": {
        "contentMediaType": "oomol/bin"
    },
    "no_used_field": "test"
}

simple_handle = {
    "handle": "test",
}

value_input_handle = {
    "handle": "test",
    "value": "test_value",
}

bin_handle = {
    "handle": "test",
    "json_schema": {
        "contentMediaType": "oomol/bin"
    },
}

var_handle = {
    "handle": "test",
    "json_schema": {
        "contentMediaType": "oomol/var"
    },
    "name": "options"
}

secret_handle = {
    "handle": "test",
    "json_schema": {
        "contentMediaType": "oomol/secret"
    }
}

serializable_var_input_handle = {
    "handle": "test",
    "json_schema": {
        "contentMediaType": "oomol/var"
    },
    "serialize_for_cache": True
}

serializable_var_output_handle = {
    "handle": "test",
    "json_schema": {
        "contentMediaType": "oomol/var"
    },
    "_serialize_for_cache": True
}

class HandleTestResult(TypedDict):
    handle: str
    description: Optional[str]
    kind: Optional[str]

    is_additional_handle: Optional[bool]
    is_var_handle: bool
    is_secret_handle: bool
    is_bin_handle: bool

class InputHandleTestResult(HandleTestResult):
    nullable: Optional[bool]
    is_serializable_var: bool
    value: Optional[Any]
    has_value: bool

class OutputHandleTestResult(HandleTestResult):
    need_serialize_var_for_cache: bool

class TestHandleData(unittest.TestCase):

    def handle_test(self, handle_def: HandleDef, expected: HandleTestResult):
        self.assertEqual(handle_def.handle, expected["handle"])
        self.assertEqual(handle_def.description, expected["description"])
        self.assertEqual(handle_def.kind, expected["kind"])
        
        self.assertEqual(handle_def.is_additional_handle(), expected["is_additional_handle"])
        self.assertEqual(handle_def.is_var_handle(), expected["is_var_handle"])
        self.assertEqual(handle_def.is_secret_handle(), expected["is_secret_handle"])
        self.assertEqual(handle_def.is_bin_handle(), expected["is_bin_handle"])

    def input_handle_test(self, input_def: InputHandleDef, expected: InputHandleTestResult):
        self.handle_test(input_def, expected)
        self.assertEqual(input_def.nullable, expected["nullable"])
        self.assertEqual(input_def.is_serializable_var(), expected["is_serializable_var"])
        self.assertEqual(input_def.value, expected["value"])
        self.assertEqual(input_def.has_value(), expected["has_value"])

    def output_handle_test(self, output_def: OutputHandleDef, expected: OutputHandleTestResult):
        self.handle_test(output_def, expected)
        self.assertEqual(output_def.need_serialize_var_for_cache(), expected["need_serialize_var_for_cache"])

    def test_error_handle(self):
        """Test that error handle raises ValueError."""
        with self.assertRaises(ValueError, msg="missing attr key: 'handle'"):
            HandleDef(**missing_handle)

    def test_simple_handle(self):
        """Test that simple handle can be created."""
        handle_def = HandleDef(**simple_handle)

        # get attributes
        self.assertEqual(handle_def.get("handle"), "test")
        self.assertEqual(handle_def["handle"], "test")
        self.assertTrue("handle" in handle_def)

        # test handle
        self.handle_test(handle_def, {
            "handle": "test",
            "description": None,
            "kind": None,
            "is_additional_handle": False,
            "is_var_handle": False,
            "is_secret_handle": False,
            "is_bin_handle": False
        })

    def test_simple_input_output_handle(self):
        # input
        self.input_handle_test(InputHandleDef(**simple_handle), {
            "handle": "test",
            "description": None,
            "kind": None,
            "nullable": None,
            "value": None,
            "is_additional_handle": False,
            "is_var_handle": False,
            "is_secret_handle": False,
            "is_bin_handle": False,
            "is_serializable_var": False,
            "has_value": False
        })

    def test_simple_output_handle(self):
        # output
        output_def = OutputHandleDef(**simple_handle)
        self.output_handle_test(output_def, {
            "handle": "test",
            "description": None,
            "kind": None,
            "is_additional_handle": False,
            "is_var_handle": False,
            "is_secret_handle": False,
            "is_bin_handle": False,
            "need_serialize_var_for_cache": False
        })

    def test_redundant_handle(self):
        """Test that redundant fields in handle are ignored."""
        handle_def = HandleDef(**redundant_handle)
        self.assertEqual(handle_def.handle, "test")

    def test_value_input_handle(self):
        self.input_handle_test(InputHandleDef(**value_input_handle), {
            "handle": "test",
            "description": None,
            "kind": None,
            "nullable": None,
            "value": "test_value",
            "is_additional_handle": False,
            "is_var_handle": False,
            "is_secret_handle": False,
            "is_bin_handle": False,
            "is_serializable_var": False,
            "has_value": True
        })

    def test_serializable_var_input_handle(self):
        """Test that serializable var input handle can be created."""
        input_def = InputHandleDef(**serializable_var_input_handle)
        self.input_handle_test(input_def, {
            "handle": "test",
            "description": None,
            "kind": None,
            "nullable": None,
            "value": None,
            "is_additional_handle": False,
            "is_var_handle": True,
            "is_secret_handle": False,
            "is_bin_handle": False,
            "is_serializable_var": True,
            "has_value": False
        })

    def test_serializable_var_output_handle(self):
        """Test that serializable var output handle can be created."""
        output_def = OutputHandleDef(**serializable_var_output_handle)
        self.output_handle_test(output_def, {
            "handle": "test",
            "description": None,
            "kind": None,
            "is_additional_handle": False,
            "is_var_handle": True,
            "is_secret_handle": False,
            "is_bin_handle": False,
            "need_serialize_var_for_cache": True
        })

    def test_var_handle(self):
        self.handle_test(HandleDef(**var_handle), {
            "handle": "test",
            "description": None,
            "kind": None,
            "is_additional_handle": False,
            "is_var_handle": True,
            "is_secret_handle": False,
            "is_bin_handle": False
        })

        self.input_handle_test(InputHandleDef(**var_handle), {
            "handle": "test",
            "description": None,
            "kind": None,
            "nullable": None,
            "value": None,
            "is_additional_handle": False,
            "is_var_handle": True,
            "is_secret_handle": False,
            "is_bin_handle": False,
            "is_serializable_var": False,
            "has_value": False
        })

        self.output_handle_test(OutputHandleDef(**var_handle), {
            "handle": "test",
            "description": None,
            "kind": None,
            "is_additional_handle": False,
            "is_var_handle": True,
            "is_secret_handle": False,
            "is_bin_handle": False,
            "need_serialize_var_for_cache": False
        })

    def test_secret_handle(self):
        self.handle_test(HandleDef(**secret_handle), {
            "handle": "test",
            "description": None,
            "kind": None,
            "is_additional_handle": False,
            "is_var_handle": False,
            "is_secret_handle": True,
            "is_bin_handle": False
        })

        self.input_handle_test(InputHandleDef(**secret_handle), {
            "handle": "test",
            "description": None,
            "kind": None,
            "nullable": None,
            "value": None,
            "is_additional_handle": False,
            "is_var_handle": False,
            "is_secret_handle": True,
            "is_bin_handle": False,
            "is_serializable_var": False,
            "has_value": False
        })

        self.output_handle_test(OutputHandleDef(**secret_handle), {
            "handle": "test",
            "description": None,
            "kind": None,
            "is_additional_handle": False,
            "is_var_handle": False,
            "is_secret_handle": True,
            "is_bin_handle": False,
            "need_serialize_var_for_cache": False
        })

    def test_bin_handle(self):
        self.handle_test(HandleDef(**bin_handle), {
            "handle": "test",
            "description": None,
            "kind": None,
            "is_additional_handle": False,
            "is_var_handle": False,
            "is_secret_handle": False,
            "is_bin_handle": True
        })

        self.input_handle_test(InputHandleDef(**bin_handle), {
            "handle": "test",
            "description": None,
            "kind": None,
            "nullable": None,
            "value": None,
            "is_additional_handle": False,
            "is_var_handle": False,
            "is_secret_handle": False,
            "is_bin_handle": True,
            "is_serializable_var": False,
            "has_value": False
        })

        self.output_handle_test(OutputHandleDef(**bin_handle), {
            "handle": "test",
            "description": None,
            "kind": None,
            "is_additional_handle": False,
            "is_var_handle": False,
            "is_secret_handle": False,
            "is_bin_handle": True,
            "need_serialize_var_for_cache": False
        })


    def test_object_handle(self):
        d = {
            "handle": "auto_slices",
            "json_schema": {
                "items": {
                    "properties": {
                        "begin": { "type": "number" },
                        "end": { "type": "number" }
                    },
                    "required": ["begin", "end"],
                    "type": "object"
                },
                "type": "array"
            }
        }
        self.handle_test(HandleDef(**d), {
            "handle": "auto_slices",
            "description": None,           
            "kind": None,
            "is_additional_handle": False,
            "is_var_handle": False,
            "is_secret_handle": False,
            "is_bin_handle": False
        })

        self.input_handle_test(InputHandleDef(**d), {
            "handle": "auto_slices",
            "description": None,
            "kind": None,
            "nullable": None,
            "value": None,
            "is_additional_handle": False,
            "is_var_handle": False,
            "is_secret_handle": False,
            "is_bin_handle": False,
            "is_serializable_var": False,
            "has_value": False
        })

        self.output_handle_test(OutputHandleDef(**d), {
            "handle": "auto_slices",
            "description": None,
            "kind": None,
            "is_additional_handle": False,
            "is_var_handle": False,
            "is_secret_handle": False,
            "is_bin_handle": False,
            "need_serialize_var_for_cache": False
        })


import time
import threading

def throttle(period):
    last_invoke_time = 0
    timer = None
    last_args = []
    last_kwargs = dict()

    def decorator(fn):
        def wrapper(*args, **kwargs):
            nonlocal last_invoke_time, timer, last_args, last_kwargs
            now = time.time()
            should_invoke = now - last_invoke_time > period

            def invoke():
                nonlocal last_invoke_time, timer, last_args, last_kwargs
                last_invoke_time = time.time()
                timer = None
                return fn(*last_args, **last_kwargs)

            if should_invoke:
                if timer:
                    timer.cancel()
                    timer = None
                last_invoke_time = now
                return fn(*args, **kwargs)
            else:
                last_args = args
                last_kwargs = kwargs
                if timer:
                    return
                timer = threading.Timer(period - (now - last_invoke_time), lambda: invoke())
                timer.start()

        return wrapper
    return decorator
